import uuid
from ruleau.constants import OverrideLevel


class Rule:
    def __init__(
        self,
        logic_func,
        name,
        depends_on,
        override_level,
        lazy_dependencies,
    ):
        """
        :param logic_func: User defined rule
        :param name: User defined human readable name of the rule
        :param depends_on: Rule dependencies
        :param override_level: Override level
        :param lazy_dependencies: Flag to switch loading of rule dependencies lazily
        """
        self.id = str(uuid.uuid4())
        self.name = name
        self.logic_func = logic_func
        self.depends_on = depends_on
        self.override_level = override_level
        self.__name__ = logic_func.__name__
        self.lazy_dependencies = lazy_dependencies

    def __str__(self) -> str:
        return self.__name__

    def __call__(self, *args, **kwargs) -> bool:
        return self.logic_func(*args, **kwargs)


def rule(
    name=None,
    depends_on=None,
    override_level=OverrideLevel.ANY_OVERRIDE,
    lazy_dependencies=False,
):
    """Decorator to encapsulate a function into a rule"""
    depends_on = depends_on or []

    def rule_decorator(func) -> Rule:
        return Rule(func, name, depends_on, override_level, lazy_dependencies)

    return rule_decorator



from typing import AnyStr, Any, List, Dict, Optional
from ruleau.execute import ExecutionResult
from urllib.parse import urljoin
import requests


class ApiAdapter:

    base_url: AnyStr
    api_key: Optional[AnyStr]
    result: ExecutionResult
    payload: Dict[AnyStr, Any]

    def __init__(
        self,
        base_url: AnyStr,
        payload: Dict[AnyStr, Any],
        result: ExecutionResult,
        api_key: Optional[AnyStr] = None,
    ):
        """
        :param base_url: Base URL of the ruleau API
        :param payload: Payload the rules executed upon
        :param result: Execution result of the ruleset
        :param api_key: (Optional) API key to authenticate with the API
        """
        self.base_url = base_url
        self.api_key = api_key
        self.result = result
        self.payload = payload

    @staticmethod
    def flatten(
        results: dict, output: list = [], order: int = 0
    ) -> List[Dict[AnyStr, Any]]:
        """Flatten the processed execution result
        :param results: Processed results of each rule
        :param output: Output to append flattened results to
        :param order: Execution Order of the rule
        :return:
        """
        if len(results["dependencies"]):
            # If there are dependencies in the list, recurse, flatten, order
            for dependency in results["dependencies"]:
                ApiAdapter.flatten(dependency, output, order + 1)
            results["dependencies"] = [x["name"] for x in results["dependencies"]]
        # Add the order to store in the API
        results["order"] = order
        # Append processed output to the final output
        output.append(results)
        return output

    @staticmethod
    def process(result: ExecutionResult) -> Dict[AnyStr, Any]:
        """Process the execution result into a dictionary
        :param result: Final processed result of the execution
        :return: A dictionary of all the data related to a case
        """
        return {
            "id": result.executed_rule.id,
            "name": result.executed_rule.name
            if result.executed_rule.name is not None
            else result.executed_rule.__name__,
            "description": result.executed_rule.__doc__,
            "dependencies": [
                ApiAdapter.process(result.dependant_results.results[rule.__name__])
                for rule in result.executed_rule.depends_on
            ],
            "result": {"result": result.value, "payload": result.payload.accessed},
        }

    def send(self) -> bool:
        """Send the case payload to the the API
        :return: boolean
        """
        rules = ApiAdapter.flatten(ApiAdapter.process(self.result), [])
        request_payload = {"payload": self.payload, "rules": rules}
        # Post the case results to the API
        response = requests.post(
            urljoin(self.base_url, "/api/v1/cases"), json=request_payload
        )
        return response.status_code == 201



from ruleau.rule import rule
from ruleau.execute import ExecutionResult
from ruleau.exceptions import TopLevelRuleRequiresNameException
from typing import AnyStr


def _validate_name(name):
    """Validator to check if top level rule has a human readable name
    :raises: TopLevelRuleRequiresNameException
    """
    if not name or not isinstance(name, str):
        raise TopLevelRuleRequiresNameException()


def Any(name: AnyStr, *args):
    """Aggregator to implement OR operation
    Returns truthy, if any one of the rule result is truthy
    """
    _validate_name(name)

    @rule(name=name, depends_on=args)
    def any_aggregator(context: ExecutionResult, _):
        return any(result.value for result in context.dependant_results)

    return any_aggregator


def All(name: AnyStr, *args):
    """Aggregator to implement AND operation
    Returns truthy, if and all of the rule results are truthy
    """
    _validate_name(name)

    @rule(name=name, depends_on=args)
    def all_aggregator(context: ExecutionResult, _):
        return all(result.value for result in context.dependant_results)

    return all_aggregator



class TopLevelRuleRequiresNameException(Exception):
    """Exception raised if top level aggregators don't have a human readable name"""

    pass


class MethodNotAllowedException(Exception):
    """Exception raised if a forbidden RuleauDict method is called"""

    pass



from os import mkdir, path
import sys
import importlib
import argparse
from argparse import Namespace
import inspect
from jinja2 import Template
from typing import Dict
import re
from ruleau.rule import Rule


def _description_for(rst: str) -> str:
    """Takes a rule docstring as ReStructuredText, and returns just the line
    representing a description
    """
    # TODO - this can be better achieved by parsing using the actual docutils rst parser
    rst = rst or ""
    if "\n\n" in rst:
        return rst[: rst.index("\n\n")]
    return rst


def _properties_for(rst: str) -> Dict[str, str]:
    """Takes a rule docstring as ReStructuredText, and returns
    a dictionary of `property name`: `property value` pairs
    """
    # TODO - this can be better achieved by parsing using the actual docutils rst parser
    properties_regex = r"\.\. prop::\s+(?P<key>[^\n]*)\s+\n\s+(?P<value>[^\n]+)"

    rst = rst or ""
    return {k: v for (k, v) in re.findall(properties_regex, rst)}


def _title_for(name: str) -> str:
    return name.replace("_", " ").title()


def generate_documentation(rules) -> str:
    """Returns a HTML string, documenting the passed in rule and its
    dependants
    """

    with open(
        path.join(path.dirname(path.realpath(__file__)), "html", "documentation.html")
    ) as f:
        doc_template = Template(f.read())
    # Add more information to rules

    def _enrich_rule(rule: Rule):
        return {
            "function_name": rule.__name__,
            "name": rule.name,
            "title": _title_for(rule.__name__),
            "override_level": rule.override_level.name,
            "sourcecode": inspect.getsource(rule.logic_func),
            "unparsed_doc": rule.logic_func.__doc__,
            "description": _description_for(rule.logic_func.__doc__),
            "doc_properties": _properties_for(rule.logic_func.__doc__),
            "dependencies": [_enrich_rule(dependent) for dependent in rule.depends_on],
        }

    enriched_rules = [_enrich_rule(rule) for rule in rules]

    return doc_template.render(rules=enriched_rules)


def render_doc_for_module(module_file) -> str:
    spec = importlib.util.spec_from_file_location(module_file, module_file)
    config = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(config)

    # Pick rules to document
    rules = [
        rule
        for rule in vars(config).values()
        if getattr(type(rule), "__name__", None) == "Rule"
    ]
    if not rules:
        raise Exception("No rules found in {}.".format(module_file))

    return generate_documentation(rules)


def generate_and_save_to_file(input_files, output_dir) -> int:
    generated_docs = []

    if len(input_files) == 0:
        raise Exception("No file(s) supplied to generate documentation for.")

    if not path.exists(output_dir):
        mkdir(output_dir)

    for target_file in input_files:
        if not path.exists(target_file):
            raise ValueError(f"{target_file} does not exist")

        documentation = render_doc_for_module(target_file)

        output_filename = path.join(
            output_dir, "{}.html".format(path.basename(target_file).split(".")[0])
        )

        with open(output_filename, "w") as f:
            f.write(documentation)
        generated_docs.append(output_filename)

    print(
        f"{len(generated_docs)} doc{'s' if len(generated_docs) > 1 else ''} generated."
    )
    print("\n".join(generated_docs))

    return 0


def get_arguments(args) -> Namespace:
    """Parses arguments for the ruleau-docs command
    :param args:
    :return:
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output-dir",
        dest="output_dir",
        default="./",
        help="Directory for generated documents",
    )
    parser.add_argument("files", nargs="*")
    return parser.parse_args(args)


def main() -> int:  # pragma: no cover
    """Console script for deft document generation.
    USAGE:
    ```bash
    $ ruleau-docs [--output-dir=<argument>] filename ...
    ```
    """
    args = get_arguments(sys.argv[1:])
    return generate_and_save_to_file(args.files, args.output_dir)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import logging
from json import dumps
from typing import Iterable, Optional, Dict
from typing import Any, AnyStr
from ruleau.constants import OverrideLevel
from copy import deepcopy
from ruleau.rule import Rule
from ruleau.structures import RuleauDict

logger = logging.getLogger(__name__)


class DependantResults:
    def __init__(self, dependants: Iterable[Rule], payload, overrides, lazy=False):
        self.dependants = {dep.__name__: dep for dep in dependants}
        self.payload = deepcopy(payload)
        self.overrides = overrides
        if not lazy:
            self.results = {
                depend.__name__: execute(depend, self.payload, self.overrides)
                for depend in dependants
            }
        else:
            self.results = {}

    def run(self, name):
        """
        Run and store the result of a rule dependency
        :param name:
        :return:
        """

        if name not in self.dependants:
            raise AttributeError(
                f"Result for rule '{name}' not available, as it was not "
                f"declared as a dependency. "
                f"depends_on={dumps(list(self.dependants.keys()))}"
            )

        if name not in self.results:
            self.results[name] = execute(
                self.dependants[name], self.payload, self.overrides
            )
        return self.results[name]

    def __getattr__(self, name):
        try:
            # Enable access to normal python properties
            return super().__getattr__(name)
        except AttributeError:
            # If an attribute wasn't found, check for the dependency
            return self.run(name)

    def __iter__(self):
        for dep in self.dependants:
            yield self.__getattr__(dep)


class ExecutionResult:
    def __init__(
        self,
        executed_rule: Rule,
        payload: RuleauDict,
        value,
        dependant_results: DependantResults,
        is_overridden: bool = False,
        original_value=None,
    ):
        self.executed_rule = executed_rule
        self.payload = payload
        self.value = value
        self.is_overridden = is_overridden
        self.original_value = original_value
        self.dependant_results = dependant_results


def execute(
    executable_rule: Rule,
    payload: Dict[AnyStr, Any],
    overrides: Optional[Dict[AnyStr, Any]] = None,
    api_config: Optional[Dict[AnyStr, Any]] = None,
) -> ExecutionResult:
    """
    Executes the provided rule, following dependencies and
    passing in results accordingly
    """
    payload_copy = RuleauDict(payload)

    depend_results = DependantResults(
        executable_rule.depends_on,
        payload,
        overrides,
        lazy=executable_rule.lazy_dependencies,
    )
    context = ExecutionResult(executable_rule, payload_copy, None, depend_results)

    result = ExecutionResult(
        executable_rule,
        payload_copy,
        executable_rule(context, payload_copy),
        depend_results,
    )

    if overrides and executable_rule.name in overrides:
        # TODO: We should check on the first call to execute that
        # the overrides specified will actually be used, i.e. it
        # is an error to pass in an override for a rule which doesn't
        # exist

        if executable_rule.override_level == OverrideLevel.NO_OVERRIDE:
            raise Exception(
                "Tried to override rule '{rule.name}' "
                "(function '{rule.__name__}'), but override level is "
                "set to NO_OVERRIDE"
            )

        override_result = overrides[executable_rule.name]

        logger.info(
            "Overriding result for rule '%s' (function '%s'): new value = '%s'",
            executable_rule.name,
            executable_rule.__name__,
            override_result,
        )

        result.is_overridden = True
        result.original_value = result.value
        result.value = override_result
    if api_config:
        from ruleau.adapter import ApiAdapter

        ApiAdapter(payload=payload, result=result, **api_config).send()
    return result



from copy import deepcopy
from ruleau.exceptions import MethodNotAllowedException
from flatdict import FlatDict
from collections.abc import MutableMapping
from json import dumps


class RuleauDict(MutableMapping):
    """
    RuleauDict is an immutable, dictionary lookalike mapping which allows the
    library to track how many times a key is used inside a payload.

    It provides a public property `.accessed` which returns all the accessed
    keys with their count and types

    >>> RuleauDict({"value": 1})
    <RuleauDict {'a': 1}>

    """

    _key: str = "__ruleau_key__"
    _value: str = "__ruleau_value__"
    _type: str = "__ruleau_type__"
    _accessed_count: str = "__ruleau_accessed_count__"

    def __init__(self, data):
        """
        Initialize RuleauDict with the dictionary passed by the user
        :param data:
        """
        super().__init__()
        self._data = {}
        self._original_data = deepcopy(data)
        self._encode(deepcopy(data))

    def __getitem__(self, key):
        """
        Accessor function to get the dictionary values and increment accessed count
        :param key:
        :return:
        """
        # If the value is accessed, increment the counter
        self._data[key][self._accessed_count] += 1
        return self._data[key][self._value]

    def __repr__(self):
        return f"<RuleauDict {dumps(self._original_data)}>"

    def __delitem__(self, key):
        raise MethodNotAllowedException()

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __setitem__(self, key, value):
        raise MethodNotAllowedException()

    def __eq__(self, other):
        return self._original_data == other

    def __ne__(self, other):
        return self._original_data != other

    def items(self):
        return self._original_data.items()

    def _encode(self, data) -> None:
        """
        Method to encode the dictionary into RuleauDict structure
        :param data:
        """
        for key, value in data.items():
            self._data[key] = {
                self._key: key,
                self._value: RuleauDict(value) if isinstance(value, dict) else value,
                self._type: type(value).__name__,
                self._accessed_count: 0,
            }

    def _decode(self):
        """
        Method to decode the RuleauDict into a nested dictionary with all metadata
        :return: Decoded dictionary with all internal values
        """
        data = {}
        for key, value in self._data.items():
            obj = deepcopy(value)
            obj.update({self._value: self._decode_value(obj[self._value])})
            data[key] = obj
        return data

    def _decode_value(self, value):
        """
        Helper function to decode the final value of the dictionary
        :param value:
        :return: Normalized decoded value
        """
        return value._decode() if isinstance(value, RuleauDict) else value

    @property
    def accessed(self):
        """
        Processes the result of the how many times a key was accessed.
        :return: Dictionary of items that were accessed with it's count
        """
        decoded_flattened = FlatDict(self._decode(), ".")
        original_flattened = FlatDict(self._original_data, ".")
        accessed_keys = [
            x
            for x in decoded_flattened
            if self._accessed_count in x and decoded_flattened[x] > 0
        ]
        accessed_values = []
        for accessed_key in accessed_keys:
            original_value_key = accessed_key.replace(
                f".{self._accessed_count}", ""
            ).replace(f".{self._value}", "")
            type_key = accessed_key.replace(self._accessed_count, self._type)
            accessed_values.append(
                {
                    "key": original_value_key,
                    "accessed_count": decoded_flattened[accessed_key],
                    "type": decoded_flattened[type_key],
                    "value": original_flattened[original_value_key].as_dict()
                    if isinstance(original_flattened[original_value_key], FlatDict)
                    else original_flattened[original_value_key],
                }
            )
        return accessed_values
